@@ -3,6 +3,8 @@ module Network.Mail.Mime
     ( -- * Datatypes
       Boundary (..)
     , Mail (..)
+    , emptyMail
+    , Address (..)
     , Alternatives
     , Part (..)
     , Encoding (..)
@@ -11,7 +13,9 @@ module Network.Mail.Mime
     , renderMail'
       -- * Sending messages
     , sendmail
+    , sendmailCustom
     , renderSendMail
+    , renderSendMailCustom
       -- * High-level 'Mail' creation
     , simpleMail
       -- * Utilities
@@ -29,7 +33,8 @@ import Control.Arrow
 import System.Process
 import System.IO
 import System.Exit
-import qualified Codec.Binary.Base64 as Base64
+import System.FilePath (takeFileName)
+import qualified Data.ByteString.Base64 as Base64
 import Control.Monad ((<=<), forM)
 import Data.List (intersperse)
 import qualified Data.Text.Lazy as LT
@@ -66,8 +71,12 @@ instance Random Boundary where
 
 -- | An entire mail message.
 data Mail = Mail
-    { -- | All headers, including to, from subject.
-      mailHeaders :: [(S.ByteString, Text)]
+    { mailFrom :: Address
+    , mailTo   :: [Address]
+    , mailCc   :: [Address]
+    , mailBcc  :: [Address]
+    -- | Other headers, excluding from, to, cc and bcc.
+    , mailHeaders :: [(S.ByteString, Text)]
     -- | A list of different sets of alternatives. As a concrete example:
     --
     -- > mailParts = [ [textVersion, htmlVersion], [attachment1], [attachment1]]
@@ -77,6 +86,23 @@ data Mail = Mail
     , mailParts :: [Alternatives]
     }
 
+-- | A mail message with the provided 'from' address and no other
+-- fields filled in.
+emptyMail :: Address -> Mail
+emptyMail from = Mail
+    { mailFrom    = from
+    , mailTo      = []
+    , mailCc      = []
+    , mailBcc     = []
+    , mailHeaders = []
+    , mailParts   = []
+    }
+
+data Address = Address
+    { addressName  :: Maybe Text
+    , addressEmail :: Text
+    }
+
 -- | How to encode a single part. You should use 'Base64' for binary data.
 data Encoding = None | Base64 | QuotedPrintableText | QuotedPrintableBinary
 
@@ -152,9 +178,10 @@ showPairs mtype parts gen =
 
 -- | Render a 'Mail' with a given 'RandomGen' for producing boundaries.
 renderMail :: RandomGen g => g -> Mail -> (L.ByteString, g)
-renderMail g0 (Mail headers parts) =
+renderMail g0 (Mail from to cc bcc headers parts) =
     (toLazyByteString builder, g'')
   where
+    addressHeaders = map showAddressHeader [("From", [from]), ("To", to), ("Cc", cc), ("Bcc", bcc)]
     pairs = map (map partToPair) parts
     (pairs', g') = helper g0 $ map (showPairs "alternative") pairs
     helper :: g -> [g -> (x, g)] -> ([x], g)
@@ -165,7 +192,8 @@ renderMail g0 (Mail headers parts) =
          in (b : bs, g__)
     ((finalHeaders, finalBuilder), g'') = showPairs "mixed" pairs' g'
     builder = mconcat
-        [ mconcat $ map showHeader headers
+        [ mconcat addressHeaders
+        , mconcat $ map showHeader headers
         , showHeader ("MIME-Version", "1.0")
         , mconcat $ map showHeader finalHeaders
         , fromByteString "\n"
@@ -176,10 +204,29 @@ showHeader :: (S.ByteString, Text) -> Builder
 showHeader (k, v) = mconcat
     [ fromByteString k
     , fromByteString ": "
-    , if needsEncodedWord v then encodedWord v else fromText v
+    , encodeIfNeeded v
     , fromByteString "\n"
     ]
 
+showAddressHeader :: (S.ByteString, [Address]) -> Builder
+showAddressHeader (k, as) =
+  if null as
+  then mempty
+  else mconcat
+    [ fromByteString k
+    , fromByteString ": "
+    , mconcat (intersperse (fromByteString ", ") . map showAddress $ as)
+    , fromByteString "\n"
+    ]
+
+showAddress :: Address -> Builder
+showAddress a = mconcat
+    [ maybe mempty ((`mappend` fromByteString " ") . encodedWord) (addressName a)
+    , fromByteString "<"
+    , fromText (addressEmail a)
+    , fromByteString ">"
+    ]
+
 showBoundPart :: Boundary -> (Headers, Builder) -> Builder
 showBoundPart (Boundary b) (headers, content) = mconcat
     [ fromByteString "--"
@@ -205,11 +252,25 @@ renderMail' m = do
     setStdGen g'
     return lbs
 
--- | Send a fully-formed email message via the sendmail executable.
+-- | Send a fully-formed email message via the default sendmail
+-- executable with default options.
 sendmail :: L.ByteString -> IO ()
-sendmail lbs = do
-    (Just hin, _, _, phandle) <- createProcess $ (proc
-        "/usr/sbin/sendmail" ["-t"]) { std_in = CreatePipe }
+sendmail = sendmailCustom "/usr/sbin/sendmail" ["-t"]
+
+-- | Render an email message and send via the default sendmail
+-- executable with default options.
+renderSendMail :: Mail -> IO ()
+renderSendMail = sendmail <=< renderMail'
+
+-- | Send a fully-formed email message via the specified sendmail
+-- executable with specified options.
+sendmailCustom :: FilePath        -- ^ sendmail executable path
+                  -> [String]     -- ^ sendmail command-line options
+                  -> L.ByteString -- ^ mail message as lazy bytestring
+                  -> IO ()
+sendmailCustom sm opts lbs = do
+    (Just hin, _, _, phandle) <- createProcess $ 
+                                 (proc sm opts) { std_in = CreatePipe }
     L.hPut hin lbs
     hClose hin
     exitCode <- waitForProcess phandle
@@ -217,9 +278,13 @@ sendmail lbs = do
         ExitSuccess -> return ()
         _ -> error $ "sendmail exited with error code " ++ show exitCode
 
--- | Render an email message and send via 'sendmail'.
-renderSendMail :: Mail -> IO ()
-renderSendMail = sendmail <=< renderMail'
+-- | Render an email message and send via the specified sendmail
+-- executable with specified options.
+renderSendMailCustom :: FilePath    -- ^ sendmail executable path
+                        -> [String] -- ^ sendmail command-line options
+                        -> Mail     -- ^ mail to render and send
+                        -> IO ()
+renderSendMailCustom sm opts = sendmailCustom sm opts <=< renderMail'
 
 -- FIXME usage of FilePath below can lead to issues with filename encoding
 
@@ -227,8 +292,8 @@ renderSendMail = sendmail <=< renderMail'
 -- alternatives and some file attachments.
 --
 -- Note that we use lazy IO for reading in the attachment contents.
-simpleMail :: Text -- ^ to
-           -> Text -- ^ from
+simpleMail :: Address -- ^ to
+           -> Address -- ^ from
            -> Text -- ^ subject
            -> LT.Text -- ^ plain body
            -> LT.Text -- ^ HTML body
@@ -239,11 +304,11 @@ simpleMail to from subject plainBody htmlBody attachments = do
         content <- L.readFile fn
         return (ct, fn, content)
     return Mail {
-        mailHeaders =
-            [ ("To", to)
-            , ("From", from)
-            , ("Subject", subject)
-            ]
+          mailFrom = from
+        , mailTo   = [to]
+        , mailCc   = []
+        , mailBcc  = []
+        , mailHeaders = [ ("Subject", subject) ]
         , mailParts =
             [ Part "text/plain; charset=utf-8" QuotedPrintableText Nothing []
             $ LT.encodeUtf8 plainBody
@@ -251,7 +316,7 @@ simpleMail to from subject plainBody htmlBody attachments = do
             $ LT.encodeUtf8 htmlBody
             ] :
             (map (\(ct, fn, content) ->
-                    [Part ct Base64 (Just $ T.pack fn) [] content]) as)
+                    [Part ct Base64 (Just $ T.pack (takeFileName fn)) [] content]) as)
         }
 
 -- | The first parameter denotes whether the input should be treated as text.
@@ -285,6 +350,12 @@ hex x
     | x < 10 = fromWord8 $ x + 48
     | otherwise = fromWord8 $ x + 55
 
+encodeIfNeeded :: Text -> Builder
+encodeIfNeeded t =
+  if needsEncodedWord t
+  then encodedWord t
+  else fromText t
+
 needsEncodedWord :: Text -> Bool
 needsEncodedWord = not . T.all isAscii
 
@@ -296,24 +367,42 @@ encodedWord t = mconcat
     ]
   where
     go front w = front `mappend` go' w
-    go' 32 = fromWord8 95
-    go' 95 = go'' 95
-    go' 63 = go'' 63
-    go' 61 = go'' 61
+    go' 32 = fromWord8 95 -- space
+    go' 95 = go'' 95 -- _
+    go' 63 = go'' 63 -- ?
+    go' 61 = go'' 61 -- =
+
+    -- The special characters from RFC 2822. Not all of these always give
+    -- problems, but at least @[];"<>, gave problems with some mail servers
+    -- when used in the 'name' part of an address.
+    go' 34 = go'' 34 -- "
+    go' 40 = go'' 40 -- (
+    go' 41 = go'' 41 -- )
+    go' 44 = go'' 44 -- ,
+    go' 46 = go'' 46 -- .
+    go' 58 = go'' 58 -- ;
+    go' 59 = go'' 59 -- ;
+    go' 60 = go'' 60 -- <
+    go' 62 = go'' 62 -- >
+    go' 64 = go'' 64 -- @
+    go' 91 = go'' 91 -- [
+    go' 92 = go'' 92 -- \
+    go' 93 = go'' 93 -- ]
     go' w
         | 33 <= w && w <= 126 = fromWord8 w
         | otherwise = go'' w
     go'' w = fromWord8 61 `mappend` hex (w `shiftR` 4)
                           `mappend` hex (w .&. 15)
 
--- Encode data into base64. Base64.encode cannot be used here
--- because it suffers from stack overflow when used with larget input.
+-- 57 bytes, when base64-encoded, becomes 76 characters.
+-- Perform the encoding 57-bytes at a time, and then append a newline.
 base64 :: L.ByteString -> Builder
-base64 = go Base64.encodeInc . groupN 10 . L.unpack
-    where
-        go encoder [] = case encoder Base64.EDone of
-            Base64.EFinal str -> fromChar8String str
-        go encoder (chunk:rest) = case encoder $ Base64.EChunk chunk of
-            Base64.EPart str next -> fromChar8String str `mappend` go next rest
-        fromChar8String = fromWriteList writeWord8 . map (toEnum . fromEnum)
-        groupN n = map (take n) . takeWhile (not . null) . iterate (drop n)
+base64 lbs
+    | L.null lbs = mempty
+    | otherwise = fromByteString x64 `mappend`
+                  fromByteString "\r\n" `mappend`
+                  base64 y
+  where
+    (x', y) = L.splitAt 57 lbs
+    x = S.concat $ L.toChunks x'
+    x64 = Base64.encode x
